
//  The VNC system is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
//  USA.
//
// If the source code for the VNC system is not available from the place 
// whence you received this file, check http://www.orl.co.uk/vnc or contact
// the authors on vnc@orl.co.uk for information on obtaining it.



// Code for IP authorization check class
// Used to allow or disallow clients based upon their IP
// Default to allowing all clients if there is an error reading
// the list file (e.g. it doesn't exist)

#include "ipauth.h"


// IP list element Constructor
vncll_IP::vncll_IP(int NewIP[IP_LENGTH]) : next(0)  {
	for(int i = 0; i < IP_LENGTH; i++)
		IP[i] = NewIP[i];
}

// IP list element destructor
vncll_IP::~vncll_IP(void)  {
	if(next)
		delete next;
}

// IP list element IsMatch check
// Just chops an IP off to match the lengh of the IP that we're 
// comparing with.  Then do a simple strcmp()
int vncll_IP::IsMatch(const int IPToCheck[IP_LENGTH])  {

	int i = 0;
	while( (i < IP_LENGTH) && (IP[i] != INVALID_IP) ) {
		if(IPToCheck[i] != IP[i])
			return 0;
		i++;
	}
	
	return 1;
}

// IP Auth constructor, defaults to empty lists (allow all)

vncIPAuth::vncIPAuth(void) : Allowed(0), Denied(0), DenyAll(false),
                             AllowedTail(0), DeniedTail(0)  {
	ReadIPListFile();
};

// IP Auth destructor

vncIPAuth::~vncIPAuth(void) {
	// ll_IP recursively deletes itself, so delete the head
	if(Allowed) 
		delete Allowed;
	if(Denied)
		delete Denied;
}


// IP Auth Read IP list file
void vncIPAuth::ReadIPListFile(void)  {

	const int	LINE_LENGTH = 100;
	char		LineBuf[LINE_LENGTH];
	int			BufPos;
	int			IPArr[IP_LENGTH] = { 0 };

	//  Open the file IP_FILENAME for reading
	FILE *ListFile = fopen(IP_FILENAME, "r");

	// See if we have a valid open file, if not
	// this is not a fatal error, the user just
	// didn't create one.  Leave without complaint.
	if(ListFile == NULL) 
		return;
	
	int AllowLength;
	int DenyLength;
	AllowLength = strlen(ALLOW);
	DenyLength  = strlen(DENY);

	while(fgets(LineBuf, LINE_LENGTH, ListFile))  {

		// Turn the Newline character to a null
		char * NewLine = strchr(LineBuf, '\xa');
		if(NewLine)
			*NewLine = '\0';

		// Find the first non-whitespace character
		BufPos = 0;
		while(isspace(LineBuf[BufPos]) && LineBuf[BufPos] != '\0') 
			BufPos++;

		if(LineBuf[BufPos] == '\0')
			continue;

		// First, see if we are allowing this IP
		if(_strnicmp(&LineBuf[BufPos], ALLOW, AllowLength) == 0)  {
			BufPos += AllowLength + 1;
			
			while(isspace(LineBuf[BufPos]) && LineBuf[BufPos] != '\0')  
				BufPos++;

			if(LineBuf[BufPos] == '\0')
				continue;

			if(_stricmp(&LineBuf[BufPos], ALL) == 0)
				continue;
			
			ProcessIP(&LineBuf[BufPos], IPArr);
			
			if(Allowed == 0)  {
				Allowed = new vncll_IP(IPArr);
				AllowedTail = Allowed;
			} else  {
				AllowedTail->next = new vncll_IP(IPArr);
				AllowedTail = AllowedTail->next;
			} // adding to the linked list
		}  else if(_strnicmp(&LineBuf[BufPos], DENY, DenyLength) == 0)  {
			BufPos += DenyLength + 1;
			
			while(isspace(LineBuf[BufPos]) && LineBuf[BufPos] != '\0')  
				BufPos++;
			
			if(LineBuf[BufPos] == '\0')
				continue;

			if(_stricmp(&LineBuf[BufPos], ALL) == 0)  {
				DenyAll = true;
				continue;
			}

			ProcessIP(&LineBuf[BufPos], IPArr);

			if(Denied == 0)  {

				Denied = new vncll_IP(IPArr);
				DeniedTail = Denied;
			
			}  else  {
				// Add on to the end of the linked list
				DeniedTail->next = new vncll_IP(IPArr);
				DeniedTail = DeniedTail->next;
			} 

		} // Decides if it is a deny/allow line

	}  // fgets() while loop

	fclose(ListFile);
}

// IP Auth decides if an IP is authorized...

int	vncIPAuth::IsAuthorizedIP(const char* pIPToCheck)  {
	
	vncll_IP*	CurrentIP;
	int		IPToCheck[IP_LENGTH];

	ProcessIP(pIPToCheck, IPToCheck);

	// first, check if we even have any allow lists
	// no allows means that we are probably unable to read the file
	// or the user doesn't understand how to allow hosts.  We default
	// to allowing all hosts in this situation
	if(Allowed == 0 && Denied == 0 && DenyAll == false)
		return 1;

	// Now, check for specifically allowed IPs.  If they are
	// explicitly listed here, we will return true immediately
	// because, following /etc/hosts.* semantics, allows are
	// always supercede denies.
	CurrentIP = Allowed;

	while(CurrentIP) {
		if(CurrentIP->IsMatch(IPToCheck))
			return 1;
		CurrentIP = CurrentIP->next;
	}

	// Now look for denied IPs.  A secure server admin will want
	// DENY ALL in his iplist.txt file, so we will check that flag
	// first.
	if(DenyAll == true)
		return 0;

	// Make sure the Denied list exists (if we have gotten this far and
	// find an empty list, the user has not specifically denied anyone).
	if(Denied == 0)
		return 1;

	// Otherwise, look for specifically denied IPs.
	CurrentIP = Denied;

	do {
		if(CurrentIP->IsMatch(IPToCheck))
			return 0;
		CurrentIP = CurrentIP->next;
	} while(CurrentIP);	

	return 1;
}

void vncIPAuth::ProcessIP(const char * RawIP, int NewIP[IP_LENGTH])  {
	const char DOT = '.';
	int IPIndex = 0;
	int i;

	for(i = 0; i < IP_LENGTH; i++)
		NewIP[i] = INVALID_NUM;

	i = 0;
	while(RawIP[i] != '\0' && IPIndex < IP_LENGTH) {
		// Get rid of the INVALID_NUM placeholders, should we come across them
		if(NewIP[IPIndex] == INVALID_NUM)
			NewIP[IPIndex] = 0;
		while(RawIP[i] != '\0' && RawIP[i] != DOT) {
			NewIP[IPIndex] = (NewIP[IPIndex] * 10) + (RawIP[i] - '0');
			i++;
		}
		if(RawIP[i] == DOT)
			i++;
		IPIndex++;
	}
}